#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <C:\Program Files (x86)\Microsoft DirectX SDK (June 2010)\Include\XInput.h>

using namespace std;

// Need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")
#pragma comment(lib, "Xinput9_1_0.lib") // Library containing necessary 360

#define DEFAULT_BUFLEN 512
#define DEFAULT_PORT "8765"
#define WIRELESS_HOST_ADDR "10.22.2.76" //mom_and_naynet --> 192.168.2.155
#define WIRED_HOST_ADDR "192.168.2.156"

#define PWM_MAX 40000 //42000 //55000 for carpet
#define BOOST_VALUE 13000 //13000
#define PWM_TURN 42000
#define DPAD_VALUE 31000
#define DEFAULT_TRIM 2100

#define SERVO_DEFAULT 150000
#define SERVO_A_OFST 0
#define SERVO_B_OFST 0x4

#define CTRL_OFST 0
#define PWM_A_OFST 0x4
#define PWM_B_OFST 0x10
#define SYNC_A_OFST 0x1C
#define SYNC_B_OFST 0x20

#define RUN_A_BIT 0
#define RUN_B_BIT 8
#define DIR_A_BIT 1
#define DIR_B_BIT 9
#define SYNC_BIT 7

#define LEFT_INPUT_DEADZONE XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE
#define RIGHT_INPUT_DEADZONE XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE
#define LEFT_X_THRESH -0.95 //.7071
#define RIGHT_X_THRESH 0.95 //.7071


struct data_packet {
	UINT32 ctrl;
	UINT32 pwmA;
	UINT32 pwmB;
	UINT32 syncA;
	UINT32 syncB;
	UINT32 servoA;
	UINT32 servoB;
	UINT32 led;
};

int __cdecl main(int argc, char **argv) 
{
    WSADATA wsaData;
    SOCKET ConnectSocket = INVALID_SOCKET;
    struct addrinfo *result = NULL,
                    *ptr = NULL,
                    hints;

    int iResult;
    int recvbuflen = DEFAULT_BUFLEN;
	struct data_packet data;
	int rc = 0;

	bool newLeftBump = false;
	bool newRightBump = false;
	bool leftBumpPressed = false;
	bool rightBumpPressed = false;
	bool done = false;

	UINT32 pwmVal = 0;
	
	XINPUT_STATE xboxState;
	DWORD dwResult;   
	float magnitude;
	float normalizedMagnitude;
	float normalizedBoost;
	char inputChar = 'a';
	char *hostAddr;
	float trimValue = DEFAULT_TRIM;

    
	data.pwmA = 0;
	data.pwmB = 0;
	data.ctrl = 0;
	data.syncA = 1;
	data.syncB = 1;
	data.servoA = SERVO_DEFAULT;
	data.servoB = SERVO_DEFAULT;
	data.led = 0;

    // Initialize Winsock
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed with error: %d\n", iResult);
        return 1;
    }

    ZeroMemory( &hints, sizeof(hints) );
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

	while (inputChar != 'y' && inputChar != 'n')
	{
		printf("Type 'y' for wireless, or 'n' for wired:\n");
		inputChar = getchar();
	}
	if (inputChar == 'n')
	{
		hostAddr = WIRED_HOST_ADDR;
	}
	else
	{
		hostAddr = WIRELESS_HOST_ADDR;
	}
    // Resolve the server address and port
    iResult = getaddrinfo(hostAddr, DEFAULT_PORT, &hints, &result);
    if ( iResult != 0 ) {
        printf("getaddrinfo failed with error: %d\n", iResult);
        WSACleanup();
        return 1;
    }

    // Attempt to connect to an address until one succeeds
    for(ptr=result; ptr != NULL ;ptr=ptr->ai_next) {

        // Create a SOCKET for connecting to server
        ConnectSocket = socket(ptr->ai_family, ptr->ai_socktype, 
            ptr->ai_protocol);
        if (ConnectSocket == INVALID_SOCKET) {
            printf("socket failed with error: %ld\n", WSAGetLastError());
            WSACleanup();
            return 1;
        }

        // Connect to server.
        iResult = connect( ConnectSocket, ptr->ai_addr, (int)ptr->ai_addrlen);
        if (iResult == SOCKET_ERROR) {
            closesocket(ConnectSocket);
            ConnectSocket = INVALID_SOCKET;
            continue;
        }
        break;
    }

    freeaddrinfo(result);

    if (ConnectSocket == INVALID_SOCKET) {
        printf("Unable to connect to server!\n");
        WSACleanup();
        return 1;
    }

	cout << "Detecting Xbox controller...   ";
	ZeroMemory( &xboxState, sizeof(XINPUT_STATE) );
	dwResult = XInputGetState( 0, &xboxState );
	while (dwResult != ERROR_SUCCESS)
	{      
        dwResult = XInputGetState( 0, &xboxState );
		Sleep(2000);
	}
	cout << "XBox Controller detected!" << endl;

	while (!done)
	{
		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER)
			newLeftBump = true;
		else
			newLeftBump = false;

		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER)
			newRightBump = true;
		else
			newRightBump = false;

		if (newLeftBump && !leftBumpPressed)
		{
			trimValue -= 100.0;
			cout << "trim value set to " << trimValue << endl;
		}
		if (newRightBump && !rightBumpPressed)
		{
			trimValue += 100.0;
			cout << "trim value set to " << trimValue << endl;
		}

		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_BACK)
		{
			data.ctrl = (1 << SYNC_BIT);
			data.pwmA = 0;
			data.pwmB = 0;
			data.syncA = 2;
			data.syncB = 2;
			data.servoA = SERVO_DEFAULT;
			data.servoB = SERVO_DEFAULT;
			data.led = 0;
			done = true;
		}
		else 
		{

				data.syncA = 1;
				data.syncB = 1;
				data.ctrl = (0 << RUN_B_BIT) | (0 << RUN_A_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT);
				data.pwmA = 0;
				data.pwmB = 0;
				
				if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_LEFT) //turn left
				{
					data.ctrl = (1 << RUN_A_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = PWM_TURN + trimValue;
						data.pwmB = PWM_TURN;
					}
					else
					{			
						data.pwmB = PWM_TURN - trimValue;
						data.pwmA = PWM_TURN;
					}

				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_RIGHT) //turn right
				{
					data.ctrl = (1 << RUN_A_BIT) | (1 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = PWM_TURN + trimValue;
						data.pwmB = PWM_TURN;
					}
					else
					{			
						data.pwmB = PWM_TURN - trimValue;
						data.pwmA = PWM_TURN;
					}
				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_DOWN) //go backwards
				{				
			   		data.ctrl = (1 << RUN_A_BIT) | (1 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = DPAD_VALUE + trimValue;
						data.pwmB = DPAD_VALUE;
					}
					else
					{			
						data.pwmB = DPAD_VALUE - trimValue;
						data.pwmA = DPAD_VALUE;
					}
					normalizedBoost = ((float) xboxState.Gamepad.bRightTrigger) / 255.0 + 1.0;
				data.pwmA = data.pwmA * normalizedBoost;
				data.pwmB = data.pwmB * normalizedBoost;
				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_UP) // go forwards
				{
					data.ctrl = (1 << RUN_A_BIT) | (0 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = DPAD_VALUE + trimValue;
						data.pwmB = DPAD_VALUE;
					}
					else
					{			
						data.pwmB = DPAD_VALUE - trimValue;
						data.pwmA = DPAD_VALUE;
					}
					normalizedBoost = ((float) xboxState.Gamepad.bRightTrigger) / 255.0 + 1.0;
				data.pwmA = data.pwmA * normalizedBoost;
				data.pwmB = data.pwmB * normalizedBoost;
				}
				
			magnitude = xboxState.Gamepad.sThumbRX;
			if ( magnitude < (-RIGHT_INPUT_DEADZONE))
			{
				magnitude += RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoB = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else if (magnitude > RIGHT_INPUT_DEADZONE)
			{
				magnitude -= RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoB = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else
			{
				data.servoB = 150000;
			}

			magnitude = xboxState.Gamepad.sThumbRY;
			if ( magnitude < (-RIGHT_INPUT_DEADZONE))
			{
				magnitude += RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoA = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else if (magnitude > RIGHT_INPUT_DEADZONE)
			{
				magnitude -= RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoA = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else
			{
				data.servoA = 150000;
			}

			data.led = 0;
		}

		iResult = send( ConnectSocket, (char *) (&data), sizeof(data), 0 );
		if (iResult == SOCKET_ERROR) {
			printf("send failed with error: %d\n", WSAGetLastError());
			closesocket(ConnectSocket);
			WSACleanup();
			return 1;
		}

			dwResult = XInputGetState( 0, &xboxState );
			if (dwResult != ERROR_SUCCESS)
			{      
				cout << "XBox controller disconnected, exiting..." << endl;
				done = true;
			}
			Sleep(50);
	}
    closesocket(ConnectSocket);
    WSACleanup();
    return 0;
}