#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include <C:\Program Files (x86)\Microsoft DirectX SDK (June 2010)\Include\XInput.h>

using namespace std;

// Need to link with Ws2_32.lib, Mswsock.lib, and Advapi32.lib
#pragma comment (lib, "Ws2_32.lib")
#pragma comment (lib, "Mswsock.lib")
#pragma comment (lib, "AdvApi32.lib")
#pragma comment(lib, "Xinput9_1_0.lib") // Library containing necessary 360

#define DEFAULT_BUFLEN 512
#define DEFAULT_PORT "8765"
#define WIRELESS_HOST_ADDR "10.22.2.76" //mom_and_naynet --> 192.168.2.155
#define WIRED_HOST_ADDR "192.168.2.156"

#define PWM_MAX 40000 //42000 //55000 for carpet
#define BOOST_VALUE 13000 //13000
#define PWM_TURN 42000
#define DPAD_VALUE 31000
#define DEFAULT_TRIM 2100

#define SERVO_DEFAULT 150000
#define SERVO_A_OFST 0
#define SERVO_B_OFST 0x4

#define CTRL_OFST 0
#define PWM_A_OFST 0x4
#define PWM_B_OFST 0x10
#define SYNC_A_OFST 0x1C
#define SYNC_B_OFST 0x20

#define RUN_A_BIT 0
#define RUN_B_BIT 8
#define DIR_A_BIT 1
#define DIR_B_BIT 9
#define SYNC_BIT 7

#define LEFT_INPUT_DEADZONE XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE
#define RIGHT_INPUT_DEADZONE XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE
#define LEFT_X_THRESH -0.95 //.7071
#define RIGHT_X_THRESH 0.95 //.7071


struct data_packet {
	UINT32 ctrl;
	UINT32 pwmA;
	UINT32 pwmB;
	UINT32 syncA;
	UINT32 syncB;
	UINT32 servoA;
	UINT32 servoB;
	UINT32 led;
};

int __cdecl main(int argc, char **argv) 
{
    WSADATA wsaData;
    SOCKET ConnectSocket = INVALID_SOCKET;
    struct addrinfo *result = NULL,
                    *ptr = NULL,
                    hints;
    //char *sendbuf = "this is a test";
    //char recvbuf[DEFAULT_BUFLEN];
	
    int iResult;
    int recvbuflen = DEFAULT_BUFLEN;
	struct data_packet data;
	int rc = 0;

	bool newLeftBump = false;
	bool newRightBump = false;
	bool leftBumpPressed = false;
	bool rightBumpPressed = false;
	bool done = false;
	//char input;
	UINT32 pwmVal = 0;
	
	XINPUT_STATE xboxState;
	DWORD dwResult;   
	float LX;
	float LY;
	float magnitude;
	float normalizedLX;
	float normalizedLY;
	float turnLX;
	float turnLY;
	float normalizedMagnitude;
	float normalizedBoost;
	bool xboxChanged;
	DWORD prevPacketNum;

	char inputChar = 'a';
	char *hostAddr;
	float trimValue = DEFAULT_TRIM;

    
	data.pwmA = 0;
	data.pwmB = 0;
	data.ctrl = 0;
	data.syncA = 1;
	data.syncB = 1;
	data.servoA = SERVO_DEFAULT;
	data.servoB = SERVO_DEFAULT;
	data.led = 0;
    // Validate the parameters
    //if (argc != 2) {
    //    printf("usage: %s server-name\n", argv[0]);
    //    return 1;
    //}

    // Initialize Winsock
    iResult = WSAStartup(MAKEWORD(2,2), &wsaData);
    if (iResult != 0) {
        printf("WSAStartup failed with error: %d\n", iResult);
        return 1;
    }

    ZeroMemory( &hints, sizeof(hints) );
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

	while (inputChar != 'y' && inputChar != 'n')
	{
		printf("Type 'y' for wireless, or 'n' for wired:\n");
		inputChar = getchar();
	}
	if (inputChar == 'n')
	{
		hostAddr = WIRED_HOST_ADDR;
	}
	else
	{
		hostAddr = WIRELESS_HOST_ADDR;
	}
    // Resolve the server address and port
    iResult = getaddrinfo(hostAddr, DEFAULT_PORT, &hints, &result);
    if ( iResult != 0 ) {
        printf("getaddrinfo failed with error: %d\n", iResult);
        WSACleanup();
        return 1;
    }

    // Attempt to connect to an address until one succeeds
    for(ptr=result; ptr != NULL ;ptr=ptr->ai_next) {

        // Create a SOCKET for connecting to server
        ConnectSocket = socket(ptr->ai_family, ptr->ai_socktype, 
            ptr->ai_protocol);
        if (ConnectSocket == INVALID_SOCKET) {
            printf("socket failed with error: %ld\n", WSAGetLastError());
            WSACleanup();
            return 1;
        }

        // Connect to server.
        iResult = connect( ConnectSocket, ptr->ai_addr, (int)ptr->ai_addrlen);
        if (iResult == SOCKET_ERROR) {
            closesocket(ConnectSocket);
            ConnectSocket = INVALID_SOCKET;
            continue;
        }
        break;
    }

    freeaddrinfo(result);

    if (ConnectSocket == INVALID_SOCKET) {
        printf("Unable to connect to server!\n");
        WSACleanup();
        return 1;
    }

	cout << "Detecting Xbox controller...   ";
	ZeroMemory( &xboxState, sizeof(XINPUT_STATE) );
	dwResult = XInputGetState( 0, &xboxState );
	while (dwResult != ERROR_SUCCESS)
	{      
        dwResult = XInputGetState( 0, &xboxState );
		Sleep(2000);
	}
	cout << "XBox Controller detected!" << endl;

	while (!done)
	{
		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER)
			newLeftBump = true;
		else
			newLeftBump = false;

		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER)
			newRightBump = true;
		else
			newRightBump = false;

		if (newLeftBump && !leftBumpPressed)
		{
			trimValue -= 100.0;
			cout << "trim value set to " << trimValue << endl;
		}
		if (newRightBump && !rightBumpPressed)
		{
			trimValue += 100.0;
			cout << "trim value set to " << trimValue << endl;
		}

		//I think these 3 lines can be removed...
		//LX = xboxState.Gamepad.sThumbLX;
		//LY = xboxState.Gamepad.sThumbLY;
		//magnitude = sqrt(LX*LX + LY*LY);

		if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_BACK)
		{
			data.ctrl = (1 << SYNC_BIT);
			data.pwmA = 0;
			data.pwmB = 0;
			data.syncA = 2;
			data.syncB = 2;
			data.servoA = SERVO_DEFAULT;
			data.servoB = SERVO_DEFAULT;
			data.led = 0;
			done = true;
		}
		else 
		{

			//LX = xboxState.Gamepad.sThumbLX;
			//LY = xboxState.Gamepad.sThumbLY;
//determine how far the controller is pushed
			//magnitude = sqrt(LX*LX + LY*LY);
			//if (magnitude > LEFT_INPUT_DEADZONE)
			//{
				//turnLX = LX / magnitude;
				//turnLY = LY / magnitude;
				//int magicValue = sqrt((magnitude*magnitude) - (tareValue * tareValue) - (2 * LX * tareValue) - (LX * LX)) - LY;
				//normalizedLX = (LX + tareValue) / magnitude;
				//normalizedLY = (LY + magicValue) / magnitude;

			  //clip the magnitude at its expected maximum value
				//if (magnitude > 32767) magnitude = 32767;
  
			  //adjust magnitude relative to the end of the dead zone
				//magnitude -= LEFT_INPUT_DEADZONE;

			  //optionally normalize the magnitude with respect to its expected range
			  //giving a magnitude value of 0.0 to 1.0
				//normalizedMagnitude = magnitude / (32767 - LEFT_INPUT_DEADZONE);

				data.syncA = 1;
				data.syncB = 1;
				//data.pwmA = (UINT32) (normalizedMagnitude * PWM_MAX);
				//data.pwmB = (UINT32) (normalizedMagnitude * PWM_MAX);
				data.ctrl = (0 << RUN_B_BIT) | (0 << RUN_A_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT);
				data.pwmA = 0;
				data.pwmB = 0;
				
				if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_LEFT) //turn left
				{
					data.ctrl = (1 << RUN_A_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = PWM_TURN + trimValue;
						data.pwmB = PWM_TURN;
					}
					else
					{			
						data.pwmB = PWM_TURN - trimValue;
						data.pwmA = PWM_TURN;
					}

				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_RIGHT) //turn right
				{
					data.ctrl = (1 << RUN_A_BIT) | (1 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = PWM_TURN + trimValue;
						data.pwmB = PWM_TURN;
					}
					else
					{			
						data.pwmB = PWM_TURN - trimValue;
						data.pwmA = PWM_TURN;
					}
				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_DOWN) //go backwards
				{				
			   		data.ctrl = (1 << RUN_A_BIT) | (1 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = DPAD_VALUE + trimValue;
						data.pwmB = DPAD_VALUE;
					}
					else
					{			
						data.pwmB = DPAD_VALUE - trimValue;
						data.pwmA = DPAD_VALUE;
					}
					normalizedBoost = ((float) xboxState.Gamepad.bRightTrigger) / 255.0 + 1.0;
				data.pwmA = data.pwmA * normalizedBoost;
				data.pwmB = data.pwmB * normalizedBoost;
				}
				else if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_DPAD_UP) // go forwards
				{
					//data.syncA = 2;
					data.ctrl = (1 << RUN_A_BIT) | (0 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
					if ( trimValue >= 0)
					{
						data.pwmA = DPAD_VALUE + trimValue;
						data.pwmB = DPAD_VALUE;
					}
					else
					{			
						data.pwmB = DPAD_VALUE - trimValue;
						data.pwmA = DPAD_VALUE;
					}
					normalizedBoost = ((float) xboxState.Gamepad.bRightTrigger) / 255.0 + 1.0;
				data.pwmA = data.pwmA * normalizedBoost;
				data.pwmB = data.pwmB * normalizedBoost;
				}
				
				//data.pwmA += normalizedBoost * BOOST_VALUE;
				//data.pwmB += normalizedBoost * BOOST_VALUE;
			//}
			/*else //if the controller is in the deadzone zero out the magnitude
			{
				data.ctrl = (0 << SYNC_BIT);
				data.pwmA = 0;
				data.pwmB = 0;
				data.syncA = 2;
				data.syncB = 2;			
			}*/
			/*magnitude = ((float) (xboxState.Gamepad.bLeftTrigger)) / 255.0;
			data.pwmB = magnitude * PWM_MAX;
			magnitude = ((float) (xboxState.Gamepad.bRightTrigger)) / 255.0;
			data.pwmA = magnitude * PWM_MAX;
			data.ctrl = (1 << RUN_A_BIT) | (0 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << RUN_B_BIT) | (0 << SYNC_BIT);
			if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_LEFT_SHOULDER)
			{
				data.ctrl &= ~(1 << DIR_B_BIT);
			}
			if (xboxState.Gamepad.wButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER)
			{
				data.ctrl |= (1 << DIR_A_BIT);
			}
			*/
			magnitude = xboxState.Gamepad.sThumbRX;
			if ( magnitude < (-RIGHT_INPUT_DEADZONE))
			{
				magnitude += RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoB = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else if (magnitude > RIGHT_INPUT_DEADZONE)
			{
				magnitude -= RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoB = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else
			{
				data.servoB = 150000;
			}

			magnitude = xboxState.Gamepad.sThumbRY;
			if ( magnitude < (-RIGHT_INPUT_DEADZONE))
			{
				magnitude += RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoA = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else if (magnitude > RIGHT_INPUT_DEADZONE)
			{
				magnitude -= RIGHT_INPUT_DEADZONE;
				normalizedMagnitude = magnitude / (32767 - RIGHT_INPUT_DEADZONE);
				data.servoA = (UINT32) (150000 - (50000 * normalizedMagnitude));
			}
			else
			{
				data.servoA = 150000;
			}

			data.led = 0;
		}

		iResult = send( ConnectSocket, (char *) (&data), sizeof(data), 0 );
		if (iResult == SOCKET_ERROR) {
			printf("send failed with error: %d\n", WSAGetLastError());
			closesocket(ConnectSocket);
			WSACleanup();
			return 1;
		}

		//xboxChanged = false;
		//while (!done && !xboxChanged)
		//{
			//prevPacketNum = xboxState.dwPacketNumber;
			//ZeroMemory( &xboxState, sizeof(XINPUT_STATE) );
			dwResult = XInputGetState( 0, &xboxState );
			if (dwResult != ERROR_SUCCESS)
			{      
				cout << "XBox controller disconnected, exiting..." << endl;
				done = true;
			}
			//else
			//{
				//if (prevPacketNum != xboxState.dwPacketNumber)
				//{
				//	xboxChanged = true;
				//}
				Sleep(50);
			//}
		//}
	}

	/*The following block implements a simplified client that accepts input with getchar, instead of the XBox controller*/
//while (!done)
//{
//   cout << "Enter a control character: ";
//   input = getchar();
//   cout << input << endl;
//   
//    switch (input)
//    {
//    case 'w':
//    	if (pwmVal < 10)
//    		pwmVal++;
//    	data.ctrl |= (0 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	break;
//    case 's':
//    	if (pwmVal > 0)
//    		pwmVal--;
//    	data.ctrl |= (0 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	break;
//    case '8':
//    	data.ctrl |= (1 << RUN_BIT) | (0 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	data.ctrl &= ~((0 << RUN_BIT) | (1 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT));
//    	data.syncA = 1;
//    	data.syncB = 1;
//    	break;
//    case '2':
//    	data.ctrl |= (1 << RUN_BIT) | (1 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	data.ctrl &= ~((0 << RUN_BIT) | (0 << DIR_A_BIT) | (1 << DIR_B_BIT) | (0 << SYNC_BIT));
//    	data.syncA = 1;
//    	data.syncB = 1;
//    	break;
//    case '4':
//    	data.ctrl |= (1 << RUN_BIT) | (1 << DIR_A_BIT) | (1 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	data.ctrl &= ~((0 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT));
//    	data.syncA = 1;
//    	data.syncB = 1;
//    	break;
//    case '6':
//    	data.ctrl |= (1 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (1 << SYNC_BIT);
//    	data.ctrl &= ~((0 << RUN_BIT) | (1 << DIR_A_BIT) | (1 << DIR_B_BIT) | (0 << SYNC_BIT));
//    	data.syncA = 1;
//    	data.syncB = 1;
//    	break;
//    case '5':
//    	data.ctrl &= ~((1 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT));
//    	data.syncA = 1;
//    	data.syncB = 1;
//    	break;
//    case 'q':
//    	data.ctrl &= ~((1 << RUN_BIT) | (0 << DIR_A_BIT) | (0 << DIR_B_BIT) | (0 << SYNC_BIT));
//      done = true;
//      break;
//    }
//    data.pwm = pwmVal * (PWM_MAX / 10);
//    cout << "writing ctrl=" << data.ctrl << " syncA=" << data.syncA << " syncB=" << data.syncB << " pwm=" << data.pwm << endl;
//    //send(socketHandle, &data, sizeof(data), 0);
//	iResult = send( ConnectSocket, (char *) (&data), sizeof(data), 0 );
//    if (iResult == SOCKET_ERROR) {
//        printf("send failed with error: %d\n", WSAGetLastError());
//        closesocket(ConnectSocket);
//        WSACleanup();
//        return 1;
//    }
//}

    // Send an initial buffer
    //iResult = send( ConnectSocket, sendbuf, (int)strlen(sendbuf), 0 );
    /*if (iResult == SOCKET_ERROR) {
        printf("send failed with error: %d\n", WSAGetLastError());
        closesocket(ConnectSocket);
        WSACleanup();
        return 1;
    }*/

    //printf("Bytes Sent: %ld\n", iResult);

    // shutdown the connection since no more data will be sent
    //iResult = shutdown(ConnectSocket, SD_SEND);
    //if (iResult == SOCKET_ERROR) {
    //    printf("shutdown failed with error: %d\n", WSAGetLastError());
    //    closesocket(ConnectSocket);
    //    WSACleanup();
    //    return 1;
    //}

    // Receive until the peer closes the connection
   /* do {

        iResult = recv(ConnectSocket, recvbuf, recvbuflen, 0);
        if ( iResult > 0 )
            printf("Bytes received: %d\n", iResult);
        else if ( iResult == 0 )
            printf("Connection closed\n");
        else
            printf("recv failed with error: %d\n", WSAGetLastError());

    } while( iResult > 0 );*/

    // cleanup
    closesocket(ConnectSocket);
    WSACleanup();

    return 0;
}